@tool
class_name RetroSelectorList extends Control

const DEFAULT_HEIGHT := 79.0
const LINE_HEIGHT_PX := 11
const LINE_HEIGHT_OFFSET := 1
const SELECTION_SMOOTHING := 8.0
const MAX_LINES := 7

const BLINK_INTERVAL := 0.2

@export var options: Array[String] = []
@export var selection_idx := 0:
	set = _set_selection_idx
@export var expanded := true:
	set(value):
		if value != expanded:
			if value:
				expand()
			else:
				select_and_contract()
			expanded = value

@export var tween_scale := 0.24

@export var debug_rebuild := false:
	set(value):
		if value:
			rebuild_list()

var _active_tween: Tween
var _scroll_line := 0

@onready var screen_bounds_container: Control = %ScreenBoundsContainer
@onready var scrolling_container: Control = %ScrollingContainer
@onready var title_list: VBoxContainer = %TitleList
@onready var selection_mask: ColorRect = %SelectionMask

@onready var scroll_arrows: Control = %ScrollArrows
@onready var up_arrow: Label = %UpArrow
@onready var down_arrow: Label = %DownArrow


func _ready() -> void:
	rebuild_list()


func _set_selection_idx(value: float) -> void:
	selection_idx = wrapi(value, 0, options.size())

	_update_scrolling()


func _update_scrolling() -> void:
	# Custom scrolling logic
	var max_line_idx := MAX_LINES - 1
	var rel_selection := selection_idx - _scroll_line
	if rel_selection < 0:
		_scroll_line = selection_idx
	if rel_selection > max_line_idx:
		_scroll_line = selection_idx - max_line_idx

	if up_arrow:
		up_arrow.visible = _scroll_line > 0
	if down_arrow:
		down_arrow.visible = options.size() - _scroll_line > MAX_LINES


## This is a workaround for an issue with the clipping boundary
func _hide_unselected() -> void:
	for i: int in title_list.get_child_count():
		if i != selection_idx:
			(title_list.get_child(i) as CanvasItem).modulate = Color.TRANSPARENT


func _show_all() -> void:
	for n: Node in title_list.get_children():
		if n is CanvasItem:
			(n as CanvasItem).modulate = Color.WHITE


func rebuild_list() -> void:
	# Remove any existing items
	for c: Node in title_list.get_children():
		c.queue_free()

	for item: String in options:
		var item_label := Label.new()
		item_label.text = item
		item_label.horizontal_alignment = HORIZONTAL_ALIGNMENT_CENTER
		title_list.add_child(item_label)

	selection_idx = (selection_idx % options.size()) if options else 0
	_update_scrolling()


func _new_tween() -> Tween:
	if _active_tween and _active_tween.is_valid():
		_active_tween.kill()
	_active_tween = create_tween()
	return _active_tween


func expand() -> Tween:
	var tween := _new_tween()
	tween.set_parallel(true)
	(
		tween
		. tween_property(self, "size:y", DEFAULT_HEIGHT, tween_scale)
		. set_trans(Tween.TRANS_CUBIC)
		. set_ease(Tween.EASE_OUT)
	)
	(
		tween
		. tween_property(screen_bounds_container, "position:y", 0, tween_scale * 1.1)
		. set_trans(Tween.TRANS_CUBIC)
		. set_ease(Tween.EASE_OUT)
	)
	_active_tween = tween
	_show_all()
	scroll_arrows.show()
	selection_mask.show()
	return tween


func select_and_contract() -> Tween:
	var contract_subtween := contract()

	var tween := create_tween()
	_active_tween = tween
	for i: int in range(3):
		tween.tween_callback(selection_mask.hide)
		tween.tween_interval(BLINK_INTERVAL * tween_scale)
		tween.tween_callback(selection_mask.show)
		tween.tween_interval(BLINK_INTERVAL * tween_scale)
	tween.tween_subtween(contract_subtween)
	return tween


func contract() -> Tween:
	var tween := _new_tween()
	var container_position := -(
		(selection_idx - _scroll_line) * LINE_HEIGHT_PX + LINE_HEIGHT_OFFSET
	)
	tween.tween_callback(scroll_arrows.hide)
	tween.set_parallel(true)
	(
		tween
		. tween_property(
			screen_bounds_container, "position:y", container_position, tween_scale * 1.1
		)
		. set_trans(Tween.TRANS_CUBIC)
		. set_ease(Tween.EASE_OUT)
	)
	(
		tween
		. tween_property(self, "size:y", LINE_HEIGHT_PX, tween_scale)
		. set_trans(Tween.TRANS_CUBIC)
		. set_ease(Tween.EASE_OUT)
	)
	tween.chain().tween_callback(_hide_unselected)
	return tween


func _process(delta: float) -> void:
	var target_position := selection_idx * LINE_HEIGHT_PX + LINE_HEIGHT_OFFSET
	selection_mask.position.y = lerpf(
		selection_mask.position.y, target_position, 1 - exp(-SELECTION_SMOOTHING * delta)
	)

	var target_scroll := -_scroll_line * LINE_HEIGHT_PX
	scrolling_container.position.y = lerpf(
		scrolling_container.position.y, target_scroll, 1 - exp(-SELECTION_SMOOTHING * delta)
	)
