@tool
class_name GunkRelay extends SignalNode
## Gunk node that rebroadcasts triggers to the network after a delay.

## Emitted when `trigger` is called, after a short delay.
signal triggered

## Emitted when `pulse` is called, after a short delay.
signal pulsed

const MAX_RUMBLE := 0.01
const MAX_QUAKE := 0.2

@export_category("Editor Tools")
@export var quick_connect_to: SignalNode:
	set = _editor_connect

@export_category("Game Scenes")
@export var editor_arrow_scene: PackedScene

var _busy := false

var _editor_arrows: Dictionary[SignalNode, EditorArrow] = {}

@onready var rumbler_3d: Rumbler3D = %Rumbler3D
@onready var quaker: Quaker = %Quaker

@onready var animation_player: AnimationPlayer = %AnimationPlayer
@onready var trigger_delay: Timer = %TriggerDelay
@onready var pulse_delay: Timer = %PulseDelay
@onready var editor_connections: Node3D = %EditorConnections


## Trigger this relay.
##
## This will emit the `triggered` signal after a delay.
func trigger() -> void:
	if not _busy:
		_busy = true
		animation_player.play("trigger")
		trigger_delay.start()


## Send a pulse through this relay.
##
## This will emit the `pulse` signal after a delay.
func pulse() -> void:
	if not _busy:
		_busy = true
		animation_player.play("pulse")
		pulse_delay.start()


func _on_trigger_delay_timeout() -> void:
	triggered.emit()
	_busy = false


func _on_pulse_delay_timeout() -> void:
	pulsed.emit()
	_busy = false


## Find a path to a target node through a GunkRelay network.
static func find_path(
	current: GunkRelay, target: GunkNode, visited: Array[GunkNode] = []
) -> Array[GunkNode]:
	# Depth-first search
	for connection: Callable in current.triggered.get_connections():
		var neighbor := connection.get_object() as GunkNode
		if neighbor and not neighbor in visited:
			if neighbor == target:
				return [current, neighbor]
			var relay := neighbor as GunkRelay
			if relay:
				var path := find_path(relay, target, visited + [current])
				if path:
					return [current] + path
	return []


func _editor_connect(node: SignalNode) -> void:
	if node:
		self.triggered.connect(node.trigger, CONNECT_PERSIST)
		self.pulsed.connect(node.pulse, CONNECT_PERSIST)
		self.notify_property_list_changed()


func _process(delta: float) -> void:
	super._process(delta)
	if not Engine.is_editor_hint():
		# Non-editor process logic
		rumbler_3d.intensity = pct_damage() * MAX_RUMBLE
		quaker.intensity = pct_damage() * MAX_QUAKE
		return

	# Check connections
	for conn: Dictionary in triggered.get_connections():
		var callable: Callable = conn.get("callable")
		if callable and callable.get_object() is SignalNode:
			var target: SignalNode = callable.get_object()
			if target != self and not target in _editor_arrows:
				var instance: EditorArrow = editor_arrow_scene.instantiate()
				instance.target = target
				editor_connections.add_child(instance)
				_editor_arrows[target] = instance

	for target: SignalNode in _editor_arrows.keys():
		if not is_instance_valid(target):
			_editor_arrows[target].queue_free()
			_editor_arrows.erase(target)
